#include "Pointing.h"

#include <RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h>

#include <armarx/control/pointing/constants/constants.h>

#include "PointingIK.h"

namespace armarx::control::pointing::core
{
    Pointing::Pointing(const Remote& remote, const Parameters& properties) :
        remote_(remote), parameters_(properties)
    {
    }

    bool
    Pointing::execute()
    {
        remote_.trajectoryPlayer->stopTrajectoryPlayer();

        // load robot
        armarx::armem::robot_state::VirtualRobotReader robotReader;
        robotReader.connect(remote_.memoryNameSystem);

        auto robot = robotReader.getSynchronizedRobot(remote_.robotName);
        if (not robot)
        {
            ARMARX_ERROR << "Could not get synchronized robot '" << remote_.robotName << "'";
            return false;
        }

        // get target
        Eigen::Vector3f target;

        auto node = robot->getRobotNode(parameters_.frameName); // TODO var name
        if (not node)
        {
            ARMARX_ERROR << "Could not get robot node '" << parameters_.frameName
                         << "'"; // TODO "robot node for frame"
        }
        target = node->toGlobalCoordinateSystemVec(parameters_.pointAt);

        ARMARX_INFO << "Pointing at " << parameters_.pointAt << " in frame '"
                    << parameters_.frameName << "'";


        visualizeTarget(target);

        auto ik = PointingIK(robot, parameters_.side);

        auto trajectory =
            ik.computeTrajectory(target); // TODO komische traj, weil wir die aus der ik nutzen

        if (not trajectory)
        {
            ARMARX_WARNING << " Target cannot be reached";
            return false;
        }

        if (trajectory->size() == 1) // TODO handshape
        {
            ARMARX_INFO << "Already pointing a target";
            return true;
        }

        trajectory = trajectory->calculateTimeOptimalTrajectory( // TODO failed mit size == 1
            0.4,
            0.3,
            0.0,
            IceUtil::Time::milliSeconds(10)); // TODO memory leak   // TODO parameter

        if (!remote_.trajectoryPlayer)
        {
            ARMARX_ERROR << "TrajectoryPlayerProxy is invalid";
        }
        remote_.trajectoryPlayer->loadJointTraj(trajectory);

        ARMARX_INFO << "Starting trajectory (" << trajectory->getTimeLength() << "s)";
        remote_.trajectoryPlayer->startTrajectoryPlayer();


        // TODO wrist position (nicht knicken)

        // TODO wrist orientation (wie gedreht, abh. von shape?)

        // TODO hands

        // TODO blocking

        return true;
    }

    void
    Pointing::visualizeTarget(Eigen::Vector3f target)
    {
        auto layer = remote_.arviz.layer(constants::ARVIZ_LAYER_NAME);
        layer.add(armarx::viz::Sphere("PointAt")
                      .radius(50)
                      .color(armarx::viz::Color::orange())
                      .position(target));
        remote_.arviz.commit(layer);
    }

} // namespace armarx::control::pointing::core
