#pragma once

#include <RobotAPI/components/ArViz/Client/Client.h>
#include <RobotAPI/interface/components/TrajectoryPlayerInterface.h>
#include <RobotAPI/libraries/armem/client/MemoryNameSystem.h>

#include "Side.h"

namespace VirtualRobot
{
    class Robot;
    using RobotPtr = std::shared_ptr<Robot>;
} // namespace VirtualRobot

namespace simox::control::method::example
{
    class PointingIK;
}

namespace armarx::control::pointing::core
{

    /**
     * @class Pointing
     * @brief Implementation of the 'Pointing'-skill.
     * @ingroup TODO
     */
    class Pointing
    {
    public:
        struct Remote
        {
            std::string robotName;
            armarx::viz::Client arviz;
            armarx::armem::client::MemoryNameSystem memoryNameSystem;
            armarx::TrajectoryPlayerInterfacePrx trajectoryPlayer;
        };

        struct Parameters
        {
            Side side;
            std::string frameName;
            Eigen::Vector3f pointAt;
            std::optional<std::string> handShape;
        };

        Pointing(const Remote&, const Parameters&);

        bool execute();

    private:
        void visualizeTarget(Eigen::Vector3f target);

    private:
        Remote remote_;
        Parameters parameters_;
    };

} // namespace armarx::control::pointing::core
