#include "Pointing.h"

#include <armarx/control/pointing/skill_ids/skill_ids.h>

#include "aron_conversions.h"

namespace armarx::control::pointing::skills
{

    armarx::skills::SkillDescription
    Pointing::GetSkillDescription()
    {
        ParamType defaultParams;
        defaultParams.frameName = "root";
        defaultParams.handShape = std::make_optional("Pointing");

        return armarx::skills::SkillDescription{.skillId = skill_ids::Pointing,
                                                .description = "Lets the robot point at something.",
                                                .rootProfileDefaults = defaultParams.toAron(),
                                                .timeout = armarx::core::time::Duration::Minutes(1),
                                                .parametersType = ParamType::ToAronType()};
    }

    Pointing::Pointing(const Remote& remote) : Base(GetSkillDescription()), remote_(remote)
    {
    }

    Pointing::MainResult
    Pointing::main(const SpecializedMainInput& in)
    {
        core::Pointing::Parameters parameters = fromAron(in.parameters);
        impl_.emplace(remote_, parameters);
        if (impl_->execute())
        {
            return MakeSucceededResult();
        }
        return MakeFailedResult();
    }

} // namespace armarx::control::pointing::skills
